from dataclasses import dataclass, field
from typing import List, Literal, Optional

from .util import SerializableDataClass


@dataclass
class RepoInfo(SerializableDataClass):
    """Information about the current HEAD of the repo."""

    commit: Optional[str] = None
    branch: Optional[str] = None
    tag: Optional[str] = None
    dirty: Optional[bool] = None
    author_name: Optional[str] = None
    author_email: Optional[str] = None
    commit_message: Optional[str] = None
    commit_time: Optional[str] = None
    git_diff: Optional[str] = None


@dataclass
class GitMetadataSettings(SerializableDataClass):
    collect: Literal["all", "some", "none"] = "all"
    fields: Optional[List[str]] = field(default_factory=list)

    @classmethod
    def merge(cls, s1: "GitMetadataSettings", s2: "GitMetadataSettings") -> "GitMetadataSettings":
        # If either is all, then return the other
        # If either is none, then return that one
        if s1.collect == "all":
            return s2
        elif s2.collect == "all":
            return s1
        elif s1.collect == "none":
            return s1
        elif s2.collect == "none":
            return s2

        assert s1.collect == "some" and s2.collect == "some"
        # intersect the fields
        ret = GitMetadataSettings(collect="some", fields=list(set(s1.fields or []).intersection(s2.fields or [])))
        if not ret.fields:
            ret.collect = "none"
        return ret



# Generic graph algorithms.

import dataclasses
from typing import Dict, List, Optional, Protocol, Set, Tuple


# An UndirectedGraph consists of a set of vertex labels and a set of edges
# between vertices.
@dataclasses.dataclass
class UndirectedGraph:
    vertices: Set[int]
    edges: Set[Tuple[int, int]]


# An AdjacencyListGraph is a mapping from vertex label to the list of vertices
# where there is a directed edge from the key to the value.
AdjacencyListGraph = Dict[int, Set[int]]


class FirstVisitF(Protocol):
    def __call__(vertex: int, *, parent_vertex: Optional[int], **kwargs) -> None:
        """Extras:
        - parent_vertex: the vertex which spawned the current vertex as its
          child during the depth-first search. `parent_vertex` is guaranteed
          to have been visited before the current one.
        """
        pass


class LastVisitF(Protocol):
    def __call__(vertex: int) -> None:
        ...


def depth_first_search(
    graph: AdjacencyListGraph,
    first_visit_f: Optional[FirstVisitF] = None,
    last_visit_f: Optional[LastVisitF] = None,
    visitation_order: Optional[List[int]] = None,
) -> None:
    """A general depth-first search algorithm over a directed graph. As it
    traverses the graph, it invokes user-provided hooks when a vertex is *first*
    visited (before visiting its children) and when it is *last* visited (after
    visiting all its children).

    The first_visit_f and last_visit_f functions may be passed additional
    information beyond the vertex being visited as kwargs. See their type
    signatures for more details. For future proofing, you will likely want to
    capture **kwargs as a catchall in your functions.

    An optional `visitation_order` can be specified, which controls the order in
    which vertices will be first visited (outside of visiting them through a
    different vertex). It can also be used to limit the set of starting vertices
    considered. Otherwise, the DFS will visit all vertices in an unspecfied
    order.
    """

    # Check the validity of the graph.
    for vs in graph.values():
        for v in vs:
            assert v in graph

    first_visited_vertices = set()
    visitation_order = visitation_order if visitation_order is not None else graph.keys()
    events = list(reversed([("first", x, dict(parent_vertex=None)) for x in visitation_order]))
    while events:
        event_type, vertex, extras = events.pop()

        if event_type == "last":
            if last_visit_f:
                last_visit_f(vertex)
            continue

        # First visit of a node. If we've already visited it, skip.
        if vertex in first_visited_vertices:
            continue
        first_visited_vertices.add(vertex)
        if first_visit_f:
            first_visit_f(vertex, parent_vertex=extras["parent_vertex"])

        # Add 'first' visitation events for all the children of the vertex to
        # the stack. But before this, add a 'last' visitation event for this
        # vertex, so that once we've completed all the children, we get the last
        # visitation event for this one.
        events.append(("last", vertex, dict()))
        for child in graph[vertex]:
            events.append(("first", child, dict(parent_vertex=vertex)))


def undirected_connected_components(graph: UndirectedGraph) -> List[List[int]]:
    """Group together all the connected components of an undirected graph.
    Return each group as a list of vertices.
    """

    # Perhaps the most performant way to implement this is via union find. But
    # in lieu of that, we can use a depth-first search over a direct-ified
    # version of the graph. Upon the first visit of each vertex, we assign it a
    # label equal to the label of the parent vertex. If there is no parent
    # vertex, we assign a new label. At the end, we can group together all the
    # vertices with the same label.

    directed_graph = {v: set() for v in graph.vertices}
    for i, j in graph.edges:
        directed_graph[i].add(j)
        directed_graph[j].add(i)

    label_counter = 0
    vertex_labels = {}

    def first_visit_f(vertex, parent_vertex, **kwargs):
        if parent_vertex is not None:
            label = vertex_labels[parent_vertex]
        else:
            nonlocal label_counter
            label = label_counter
            label_counter += 1
        vertex_labels[vertex] = label

    depth_first_search(directed_graph, first_visit_f=first_visit_f)
    output = [[] for _ in range(label_counter)]
    for vertex, label in vertex_labels.items():
        output[label].append(vertex)

    return output


def topological_sort(graph: AdjacencyListGraph, visitation_order: Optional[List[int]] = None) -> List[int]:
    """The topological_sort function accepts a graph as input, with edges from
    parents to children. It returns an ordering where parents are guaranteed to
    come before their children.

    The `visitation_order` is forwarded directly to `depth_first_search`.

    Ordering with respect to cycles is unspecified. It is the caller's
    responsibility to check for cycles if it matters.
    """

    # We use DFS, where upon the 'last' visitation of a node, we append it to
    # the final ordering. Then we reverse the list at the end.
    reverse_ordering = []

    def last_visit_f(vertex, **kwargs):
        reverse_ordering.append(vertex)

    depth_first_search(graph, last_visit_f=last_visit_f, visitation_order=visitation_order)
    reverse_ordering.reverse()
    return reverse_ordering



from dataclasses import dataclass
from typing import Dict, List, Literal, Optional, Union

from .util import SerializableDataClass


@dataclass
class PromptCompletionBlock(SerializableDataClass):
    content: str
    type: Literal["completion"] = "completion"


@dataclass
class FunctionCall(SerializableDataClass):
    name: str
    arguments: str


@dataclass
class ToolCall(SerializableDataClass):
    id: str
    function: FunctionCall
    type: Literal["function"] = "function"


@dataclass
class TextPart(SerializableDataClass):
    text: str
    type: Literal["text"] = "text"


@dataclass
class ImageURL(SerializableDataClass):
    url: str
    detail: Literal["auto", "low", "high"] = "auto"


@dataclass
class ImagePart(SerializableDataClass):
    image_url: ImageURL
    type: Literal["image_url"] = "image_url"


@dataclass
class PromptMessage(SerializableDataClass):
    content: Union[str, List[Union[TextPart, ImagePart]]]
    role: Literal["system", "user", "assistant", "function", "tool", "model"]
    name: str = None
    function_call: Union[str, FunctionCall] = None
    tool_calls: Optional[List[ToolCall]] = None


@dataclass
class PromptChatBlock(SerializableDataClass):
    messages: List[PromptMessage]
    tools: Optional[str] = None
    type: Literal["chat"] = "chat"


PromptBlockData = Union[PromptCompletionBlock, PromptChatBlock]


@dataclass
class PromptData(SerializableDataClass):
    prompt: Optional[PromptBlockData] = None
    options: Optional[Dict] = None


@dataclass
class PromptSchema(SerializableDataClass):
    id: str
    project_id: str
    _xact_id: str
    name: str
    slug: str
    description: Optional[str]
    prompt_data: PromptData
    tags: Optional[List[str]]


BRAINTRUST_PARAMS = ["use_cache"]



# Serialization format for capturing all relevant information about a span
# necessary for distributed logging / tracing. Meant to be passed around as an
# opaque string.

import base64
import dataclasses
from enum import Enum, auto
from typing import Optional
from uuid import UUID


def _try_make_uuid(s):
    try:
        ret = UUID(s).bytes
        assert len(ret) == 16
        return ret, True
    except Exception:
        return s.encode("utf-8"), False


ENCODING_VERSION_NUMBER = 1

INVALID_ENCODING_ERRMSG = "SpanComponents string is not properly encoded. This may be due to a version mismatch between the SDK library used to export the span and the library used to decode it. Please make sure you are using the same SDK version across the board"


class SpanObjectType(Enum):
    EXPERIMENT = auto()
    PROJECT_LOGS = auto()

    def __str__(self):
        return {SpanObjectType.EXPERIMENT: "experiment", SpanObjectType.PROJECT_LOGS: "project_logs"}[self]


@dataclasses.dataclass
class SpanRowIds:
    row_id: str
    span_id: str
    root_span_id: str

    def __post_init__(self):
        assert isinstance(self.row_id, str)
        assert isinstance(self.span_id, str)
        assert isinstance(self.root_span_id, str)
        assert self.row_id
        assert self.span_id
        assert self.root_span_id


@dataclasses.dataclass
class SpanComponents:
    object_type: SpanObjectType
    object_id: str
    row_ids: Optional[SpanRowIds] = None

    def __post_init__(self):
        assert isinstance(self.object_type, SpanObjectType)
        assert isinstance(self.object_id, str)
        if self.row_ids is not None:
            assert isinstance(self.row_ids, SpanRowIds)

    def to_str(self) -> str:
        # Our binary object format is as follows:
        #   - Byte 0 encodes the version number of the encoded string. This is
        #   used to check for incompatibilities with previous iterations.
        #   - Byte 1 encodes the SpanObjectType.
        #   - Byte 2 (0 or 1) describes whether or not the (row_id,
        #   span_id, root_span_id) triple is present.
        #   - Byte 3 (0 or 1) describes whether or not the row_id component
        #   is a UUID. If not, it is assumed to be a utf-8 encoded string.
        #   - Bytes 4-19 encode the object_id as a UUID
        #   - If the row triple is present, bytes 20-51 encode the span_id +
        #   root_span_id as a UUID.
        #   - If the row triple is present, the remaining bytes encode the
        #   row_id either as UUID or as UTF-8.

        if self.row_ids:
            row_id_bytes, row_id_is_uuid = _try_make_uuid(self.row_ids.row_id)
        else:
            row_id_bytes, row_id_is_uuid = None, False

        raw_bytes = bytes(
            [
                ENCODING_VERSION_NUMBER,
                self.object_type.value,
                1 if self.row_ids else 0,
                1 if row_id_is_uuid else 0,
            ]
        )

        object_id_bytes, object_id_is_uuid = _try_make_uuid(self.object_id)
        if not object_id_is_uuid:
            raise Exception("object_id component must be a valid UUID")
        raw_bytes += object_id_bytes

        if self.row_ids:
            span_id_bytes, span_id_is_uuid = _try_make_uuid(self.row_ids.span_id)
            if not span_id_is_uuid:
                raise Exception("span_id component must be a valid UUID")
            root_span_id_bytes, root_span_id_is_uuid = _try_make_uuid(self.row_ids.root_span_id)
            if not root_span_id_is_uuid:
                raise Exception("root_span_id component must be a valid UUID")
            raw_bytes += span_id_bytes
            raw_bytes += root_span_id_bytes
            raw_bytes += row_id_bytes

        return base64.b64encode(raw_bytes).decode()

    @staticmethod
    def from_str(s: str) -> "SpanComponents":
        try:
            raw_bytes = base64.b64decode(s.encode())
            assert raw_bytes[0] == ENCODING_VERSION_NUMBER
            object_type = SpanObjectType(raw_bytes[1])
            assert raw_bytes[2] in [0, 1]
            assert raw_bytes[3] in [0, 1]
            has_row_id = raw_bytes[2] == 1
            row_id_is_uuid = raw_bytes[3] == 1

            object_id = str(UUID(bytes=raw_bytes[4:20]))
            if has_row_id:
                span_id = str(UUID(bytes=raw_bytes[20:36]))
                root_span_id = str(UUID(bytes=raw_bytes[36:52]))
                if row_id_is_uuid:
                    row_id = str(UUID(bytes=raw_bytes[52:]))
                else:
                    row_id = raw_bytes[52:].decode("utf-8")
                row_ids = SpanRowIds(row_id=row_id, span_id=span_id, root_span_id=root_span_id)
            else:
                row_ids = None

            return SpanComponents(object_type=object_type, object_id=object_id, row_ids=row_ids)
        except Exception:
            raise Exception(INVALID_ENCODING_ERRMSG)

    def object_id_fields(self):
        if self.object_type == SpanObjectType.EXPERIMENT:
            return dict(experiment_id=self.object_id)
        elif self.object_type == SpanObjectType.PROJECT_LOGS:
            return dict(project_id=self.object_id, log_id="g")



from typing import Dict, List, Optional

from .db_fields import IS_MERGE_FIELD, PARENT_ID_FIELD
from .graph_util import UndirectedGraph, topological_sort, undirected_connected_components
from .util import merge_dicts


def _generate_merged_row_key(row, use_parent_id_for_id=False):
    return tuple(
        row.get(k)
        for k in [
            "org_id",
            "project_id",
            "experiment_id",
            "dataset_id",
            "prompt_session_id",
            "log_id",
            PARENT_ID_FIELD if use_parent_id_for_id else "id",
        ]
    )


def merge_row_batch(rows: List[Dict]) -> List[List[Dict]]:
    """Given a batch of rows, merges conflicting rows together to end up with a
    set of rows to insert. Returns a set of de-conflicted rows, as a list of
    lists, where separate lists contain "independent" rows which can be
    processed concurrently, while the rows in each list must be processed in
    order, as later rows may depend on earlier ones.

    Note that the returned rows will be the same objects as the input `rows`,
    meaning they are mutated in place.

    There are a few important considerations for the merge procedure:

    - Ensuring we only log one version of each row to the DB:

    Imagine we have several rows in the batch with the same ID:

        [{"make_object_id(..)": "xyz", "id": 1, "value": 1},
         {"make_object_id(...)": "xyz", "id": 1, "value": 2}]

    If we log both rows and assign them both the same transaction ID, future
    queries will not be able to disamgiguate ordering here (i.e that `value: 2`
    is the "later" value). So we must consolidate these rows into one before
    logging.

    - Merging rows with IS_MERGE_FIELD == True:

    Rows can either be incrementally updated or replaced entirely. For a
    particular row, we use the IS_MERGE_FIELD to determine whether we merge or
    replace. In case there are several incremental updates to the same row
    within the batch, we merge them into one incremental update here, so that we
    only need to do one merge in the DB.

    We need to be careful to preserve the correct value of IS_MERGE_FIELD with
    respect to the DB. For instance, if we have one batch of rows:

        [{"make_object_id(...)": "xyz", "id": 1, "value": {"a": 12}},
         {"make_object_id(...)": "xyz", "id": 1, "value": {"b": 13},
          IS_MERGE_FIELD: True}]

    We need to make sure the row inserted into the DB has IS_MERGE_FIELD == False,
    otherwise we might merge it with a previous existing version of the row.
    """

    # Check that no row is missing an ID.
    for row in rows:
        if row.get("id") is None:
            raise Exception(
                "Logged row is missing an id. This is an internal braintrust error. Please contact us at info@braintrust.dev for help"
            )

    row_groups = {}
    for row in rows:
        key = _generate_merged_row_key(row)
        existing_row = row_groups.get(key)
        # If there is an existing row and the new row has the IS_MERGE_FIELD ==
        # True property, we merge it with the existing row. Otherwise we can
        # replace it.
        if existing_row is not None and row.get(IS_MERGE_FIELD):
            # Preserve IS_MERGE_FIELD == False if the existing_row had it set to
            # false.
            preserve_nomerge = not existing_row.get(IS_MERGE_FIELD)
            merge_dicts(existing_row, row)
            if preserve_nomerge:
                del existing_row[IS_MERGE_FIELD]
        else:
            row_groups[key] = row

    merged = list(row_groups.values())

    # Now that we have just one row per id, we can bucket and order the rows by
    # their PARENT_ID_FIELD relationships.
    row_to_label = {_generate_merged_row_key(r): i for i, r in enumerate(merged)}

    # Form a graph where edges go from parents to their children.
    graph = {i: set() for i in range(len(merged))}
    for i, r in enumerate(merged):
        parent_id = r.get(PARENT_ID_FIELD)
        if not parent_id:
            continue
        parent_row_key = _generate_merged_row_key(r, use_parent_id_for_id=True)
        parent_label = row_to_label.get(parent_row_key)
        if parent_label is not None:
            graph[parent_label].add(i)

    # Group together all the connected components of the undirected graph to get
    # all groups of rows which each row in a group has a PARENT_ID_FIELD
    # relationship with at least one other row in the group.
    connected_components = undirected_connected_components(
        UndirectedGraph(vertices=set(graph.keys()), edges=set((k, v) for k, vs in graph.items() for v in vs))
    )

    # For each connected row group, run topological sort over that subgraph to
    # get an ordering of rows where parents come before children.
    buckets = [topological_sort(graph, visitation_order=cc) for cc in connected_components]
    return [[merged[i] for i in bucket] for bucket in buckets]


def batch_items(
    items: List[List[str]], batch_max_num_items: Optional[int] = None, batch_max_num_bytes: Optional[int] = None
) -> List[List[List[str]]]:
    """Repartition the given list of items into sets of batches which can be
    published in parallel or in sequence.

    Output-wise, each outer List[List[str]] is a set of batches which must be
    published in sequence. Within each set of batches, each individual List[str]
    batch may be published in parallel with all other batches in its set,
    retaining the order within the batch. So from outside to inside, it goes
    ordered -> parallel -> ordered.

    Arguments:

    - `items` is a list of ordered buckets, where the constraint is that items
      in different buckets can be published in parallel, while items within a
      bucket must be published in sequence. That means that if two items are in
      the same bucket, they will either appear in the same innermost List[str]
      in the output, or in separate List[List[str]] batch sets, with their
      relative order preserved. If two items are in different buckets, they can
      appear in different List[str] batches.

    - `batch_max_num_items` is the maximum number of items in each List[str]
      batch. If not provided, there is no limit on the number of items.

    - `batch_max_num_bytes` is the maximum number of bytes (computed as
      `sum(len(item) for item in batch)`) in each List[str] batch. If an
      individual item exceeds `batch_max_num_bytes` in size, we will place it in
      its own batch. If not provided, there is no limit on the number of bytes.
    """

    if batch_max_num_items is None:
        batch_max_num_items = float("inf")
    if batch_max_num_bytes is None:
        batch_max_num_bytes = float("inf")

    assert batch_max_num_items > 0

    output = []
    next_items = []
    batch_set = []
    batch = []
    batch_len = 0

    def add_to_batch(item):
        nonlocal batch_len
        batch.append(item)
        batch_len += len(item)

    def flush_batch():
        nonlocal batch, batch_len
        batch_set.append(batch)
        batch = []
        batch_len = 0

    while items:
        for bucket in items:
            i = 0
            for item in bucket:
                if len(batch) == 0 or (
                    len(item) + batch_len < batch_max_num_bytes and len(batch) < batch_max_num_items
                ):
                    add_to_batch(item)
                elif i == 0:
                    # If the very first item in the bucket fills the batch, we
                    # can flush this batch and start a new one which includes
                    # this item.
                    flush_batch()
                    add_to_batch(item)
                else:
                    break
                i += 1
            # If we didn't completely exhaust the bucket, save it for the next
            # batch set.
            if i < len(bucket):
                next_items.append(bucket[i:])
            # If we have filled the batch, flush it.
            if batch_len >= batch_max_num_bytes or len(batch) >= batch_max_num_items:
                flush_batch()

        # We've finished an iteration through all the buckets. Anything
        # remaining in `next_items` will need to be processed in a subsequent
        # batch set, so flush our remaining batch and the batch set, and use
        # next_items for the next iteration.
        if batch:
            flush_batch()
        if batch_set:
            output.append(batch_set)
            batch_set = []
        items, next_items = next_items, []

    return output



import dataclasses
import json
import sys
import urllib.parse
from typing import Dict, List, Set, Tuple, Union, get_origin


# Taken from
# https://stackoverflow.com/questions/5574702/how-do-i-print-to-stderr-in-python.
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class SerializableDataClass:
    def as_dict(self):
        """Serialize the object to a dictionary."""
        return dataclasses.asdict(self)

    def as_json(self, **kwargs):
        """Serialize the object to JSON."""
        return json.dumps(self.as_dict(), **kwargs)

    @classmethod
    def from_dict(cls, d: Dict):
        """Deserialize the object from a dictionary. This method
        is shallow and will not call from_dict() on nested objects."""
        fields = set(f.name for f in dataclasses.fields(cls))
        filtered = {k: v for k, v in d.items() if k in fields}
        return cls(**filtered)

    @classmethod
    def from_dict_deep(cls, d: Dict):
        """Deserialize the object from a dictionary. This method
        is deep and will call from_dict_deep() on nested objects."""
        fields = {f.name: f for f in dataclasses.fields(cls)}
        filtered = {}
        for k, v in d.items():
            if k not in fields:
                continue

            if (
                isinstance(v, dict)
                and isinstance(fields[k].type, type)
                and issubclass(fields[k].type, SerializableDataClass)
            ):
                filtered[k] = fields[k].type.from_dict_deep(v)
            elif get_origin(fields[k].type) == Union:
                for t in fields[k].type.__args__:
                    if isinstance(t, type) and issubclass(t, SerializableDataClass):
                        try:
                            filtered[k] = t.from_dict_deep(v)
                            break
                        except TypeError:
                            pass
                else:
                    filtered[k] = v
            elif (
                isinstance(v, list)
                and get_origin(fields[k].type) == list
                and len(fields[k].type.__args__) == 1
                and isinstance(fields[k].type.__args__[0], type)
                and issubclass(fields[k].type.__args__[0], SerializableDataClass)
            ):
                filtered[k] = [fields[k].type.__args__[0].from_dict_deep(i) for i in v]
            else:
                filtered[k] = v
        return cls(**filtered)


def coalesce(*args):
    """Returns the first non-None value in the list of `args`, or `None` if they
    are all `None`.
    """

    for a in args:
        if a is not None:
            return a
    return None


def merge_dicts_with_paths(merge_into: Dict, merge_from: Dict, path: Tuple[str], merge_paths: Set[Tuple[str]]):
    """Merges merge_from into merge_into, destructively updating merge_into. Does not merge any further than
    merge_paths."""

    if not isinstance(merge_into, dict):
        raise ValueError("merge_into must be a dictionary")
    if not isinstance(merge_from, dict):
        raise ValueError("merge_from must be a dictionary")

    for k, merge_from_v in merge_from.items():
        full_path = path + (k,)
        merge_into_v = merge_into.get(k)
        if isinstance(merge_into_v, dict) and isinstance(merge_from_v, dict) and full_path not in merge_paths:
            merge_dicts_with_paths(merge_into_v, merge_from_v, full_path, merge_paths)
        else:
            merge_into[k] = merge_from_v

    return merge_into


def merge_dicts(merge_into: Dict, merge_from: Dict):
    """Merges merge_from into merge_into, destructively updating merge_into."""

    return merge_dicts_with_paths(merge_into, merge_from, (), set())


def encode_uri_component(name):
    """Encode a single component of a URI. Slashes are encoded as well, so this
    should not be used for multiple slash-separated URI components."""

    return urllib.parse.quote(name, safe="")



TOP_BITS = 0x0DE1 << 48

MOD = 2**64
COPRIME = 205891132094649
COPRIME_INVERSE = 1522336535492693385


def modular_multiply(value: int, prime: int):
    return (value * prime) % MOD


# value : int | str
# Cannot use a | because of python 3.8
def prettify_xact(value) -> str:
    encoded = modular_multiply(int(value), COPRIME)
    return hex(encoded)[2:].rjust(16, "0")


def load_pretty_xact(encoded_hex: str) -> str:
    value = int(encoded_hex, 16)
    multiplied_inverse = modular_multiply(value, COPRIME_INVERSE)
    with_top_bits = TOP_BITS | multiplied_inverse
    return str(with_top_bits)



TRANSACTION_ID_FIELD = "_xact_id"
OBJECT_DELETE_FIELD = "_object_delete"
CREATED_FIELD = "created"
ID_FIELD = "id"

IS_MERGE_FIELD = "_is_merge"
MERGE_PATHS_FIELD = "_merge_paths"

AUDIT_SOURCE_FIELD = "_audit_source"
AUDIT_METADATA_FIELD = "_audit_metadata"
VALID_SOURCES = ["app", "api", "external"]

PARENT_ID_FIELD = "_parent_id"

ALL_ROW_ID_FIELDS = [
    ID_FIELD,
    TRANSACTION_ID_FIELD,
    CREATED_FIELD,
    IS_MERGE_FIELD,
    MERGE_PATHS_FIELD,
    OBJECT_DELETE_FIELD,
]



import dataclasses
import json


class BraintrustJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o) and not isinstance(o, type):
            return dataclasses.asdict(o)

        # Attempt to dump a Pydantic v2 `BaseModel`.
        try:
            return o.model_dump()
        except (AttributeError, TypeError):
            pass

        # Attempt to dump a Pydantic v1 `BaseModel`.
        try:
            return o.dict()
        except (AttributeError, TypeError):
            pass

        # When everything fails, just return the string representation of the object
        return str(o)


def bt_dumps(obj, **kwargs) -> str:
    return json.dumps(obj, cls=BraintrustJSONEncoder, allow_nan=False, **kwargs)



import dataclasses
from abc import ABC, abstractmethod
from typing import Dict, Optional

from .util import SerializableDataClass, eprint


@dataclasses.dataclass
class Score(SerializableDataClass):
    """A score for an evaluation. The score is a float between 0 and 1."""

    name: str
    """The name of the score. This should be a unique name for the scorer."""

    score: Optional[float]
    """The score for the evaluation. This should be a float between 0 and 1. If the score is None, the evaluation is considered to be skipped."""

    metadata: Dict[str, any] = dataclasses.field(default_factory=dict)
    """Metadata for the score. This can be used to store additional information about the score."""

    # DEPRECATION_NOTICE: this field is deprecated, as errors are propagated up to the caller.
    error: Exception = None
    """Deprecated: The error field is deprecated, as errors are now propagated to the caller. The field will be removed in a future version of the library."""

    def as_dict(self):
        return {
            "score": self.score,
            "metadata": self.metadata,
        }

    def __post_init__(self):
        if self.score is not None and (self.score < 0 or self.score > 1):
            raise ValueError(f"score ({self.score}) must be between 0 and 1")
        if self.error is not None:
            eprint(
                "The error field is deprecated, as errors are now propagated to the caller. The field will be removed in a future version of the library"
            )


class Scorer(ABC):
    async def eval_async(self, output, expected=None, **kwargs):
        return await self._run_eval_async(output, expected, **kwargs)

    def eval(self, output, expected=None, **kwargs):
        return self._run_eval_sync(output, expected, **kwargs)

    def __call__(self, output, expected=None, **kwargs):
        return self.eval(output, expected, **kwargs)

    async def _run_eval_async(self, output, expected=None, **kwargs) -> Score:
        # By default we just run the sync version in a thread
        return self._run_eval_sync(output, expected, **kwargs)

    def _name(self) -> str:
        return self.__class__.__name__

    @abstractmethod
    def _run_eval_sync(self, output, expected=None, **kwargs) -> Score:
        ...


__all__ = ["Score", "Scorer"]



import json
import os

import setuptools

dir_name = os.path.abspath(os.path.dirname(__file__))

version_contents = {}
with open(os.path.join(dir_name, "src", "braintrust_core", "version.py"), encoding="utf-8") as f:
    exec(f.read(), version_contents)

with open(os.path.join(dir_name, "README.md"), "r", encoding="utf-8") as f:
    long_description = f.read()

install_requires = []

extras_require = {}

setuptools.setup(
    name="braintrust_core",
    version=version_contents["VERSION"],
    author="Braintrust",
    author_email="info@braintrust.dev",
    description="Shared core dependencies for Braintrust packages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.braintrust.dev",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8.0",
    install_requires=install_requires,
    extras_require=extras_require,
)



from .db_fields import MERGE_PATHS_FIELD

DEFAULT_IS_LEGACY_DATASET = True


def ensure_dataset_record(r, legacy: bool):
    if legacy:
        return ensure_legacy_dataset_record(r)
    else:
        return ensure_new_dataset_record(r)


def ensure_legacy_dataset_record(r):
    if "output" in r:
        return r
    row = {**r}
    row["output"] = row.pop("expected")
    return row


def ensure_new_dataset_record(r):
    if "expected" in r:
        return r
    row = {**r}
    row["expected"] = row.pop("output")
    return row


def make_legacy_event(e):
    if "dataset_id" not in e or "expected" not in e:
        return e

    event = {**e}
    event["output"] = event.pop("expected")

    if MERGE_PATHS_FIELD in event:
        for path in event[MERGE_PATHS_FIELD] or []:
            if len(path) > 0 and path[0] == "expected":
                path[0] = "output"

    return event
